@@ -24,6 +24,30 @@ for to log, configure and control the run-time of your GUI app via command line
 :ref:`application flow`, explained further down.
 
 
+application events
+------------------
+
+The following application events are fired exactly one time in the lifetime of an app instance in the following order:
+
+* `on_app_init`: fired **after** :class:`ConsoleApp`-parts of the app instance got initialized (detected config files)
+  and **before** the image and sound resources and app states get loaded and the GUI framework app class instance gets
+  initialized.
+* `on_app_start`: fired **after** the call of the method :meth:`~MainAppBase.run_app` and **before** the command line
+  options got parsed.
+* `on_app_run`: fired **after** the parsing of the command line arguments and options and **before** the framework event
+  loop of the used GUI framework gets started.
+* `on_app_stop`: fired **after** the call of the method :meth:`~MainAppBase.stop_app` and **before** the framework win
+  will be closed.
+* `on_app_exit`: fired **after* framework win got closed and just **before** the event loop of the GUI framework will be
+  stopped and the app shutdown.
+* `on_app_quit`: fired **after** the event loop of the GUI framework got stopped and before the :meth:`AppBase.shutdown`
+  method will be called.
+
+.. note::
+    The `on_app_stop` and `on_app_exit` events will only be fired if the app is explicitly calling the
+    :meth:`~MainAppBase.stop_app` method.
+
+
 application status
 ------------------
 
@@ -102,6 +126,27 @@ to the app states data store and for the app state config variables :attr:`~Main
     * :data:`THEME_DARK_FONT_COLOR`
 
 
+app state change and save events
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+
+Notification events get fired when one of the app state variables gets changed - like e.g. the :ref:`application flow`
+or if the user changes the font size.
+
+The method name of the app state change notification event consists of the prefix ``on_`` followed by the variable name
+of the app state. So e.g. on a change of the `font_size` app state the notification event `on_font_size` will be
+fired/called (if exists as a method of the main app instance). These events don't provide any event arguments.
+
+Another event gets fired for each app state value just after the app states getting retrieved from the app class
+instance and before they get stored into the main config file.
+
+This method name of this event includes the key/name of the app state using the format
+:meth:`on_app_state_<app_state_key>_save`, so e.g. for the app state `flow_id` the method name will result in
+:meth:`on_app_state_flow_id_save`.
+
+The event is taking one event argument with the value of the app state and if it returns a value that is
+not `None` then this value will be stored/saved.
+
+
 application flow
 ----------------
 
@@ -130,70 +175,39 @@ For to go back to a previous flow in the flow path call :meth:`~MainBaseApp.chan
 defined in :data:`ACTIONS_REDUCING_FLOW_PATH`.
 
 
-application events
-------------------
-
-The following application events are fired exactly one time in the lifetime of an app instance in the following order:
-
-* `on_app_init`: fired **after** :class:`ConsoleApp`-parts of the app instance got initialized (detected config files)
-  and **before** the image and sound resources and app states get loaded and the GUI framework app class instance gets
-  initialized.
-* `on_app_start`: fired **after** the call of the method :meth:`~MainAppBase.run_app` and **before** the command line
-  options got parsed.
-* `on_app_run`: fired **after** the parsing of the command line arguments and options and **before** the framework event
-  loop of the used GUI framework gets started.
-* `on_app_stop`: fired **after** the call of the method :meth:`~MainAppBase.stop_app` and **before** the framework win
-  will be closed.
-* `on_app_exit`: fired **after* framework win got closed and just **before** the event loop of the GUI framework will be
-  stopped and the app shutdown.
-* `on_app_quit`: fired **after** the event loop of the GUI framework got stopped and before the :meth:`AppBase.shutdown`
-  method will be called.
-
-.. note::
-    The `on_app_stop` and `on_app_exit` events will only be fired if the app is explicitly calling the
-    :meth:`~MainAppBase.stop_app` method.
-
-
-application state change and save events
-----------------------------------------
-
-Notification events get fired when one of the app state variables gets changed - like e.g. the :ref:`application flow`
-or if the user changes the font size.
-
-The method name of the app state change notification event consists of the prefix ``on_`` followed by the variable name
-of the app state. So e.g. on a change of the `font_size` app state the notification event `on_font_size` will be
-fired/called (if exists as a method of the main app instance). These events don't provide any event arguments.
-
-Another event gets fired for each app state value just after the app states getting retrieved from the app class
-instance and before they get stored into the main config file.
-
-This method name of this event includes the key/name of the app state using the format
-:meth:`on_app_state_<app_state_key>_save`, so e.g. for the app state `flow_id` the method name will result in
-:meth:`on_app_state_flow_id_save`.
-
-The event is taking one event argument with the value of the app state and if it returns a value that is not `None` then
-this value will be stored/saved.
-
-
 application flow change events
-------------------------------
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
-The name of the event method that gets fired when the app want to change the flow (via the method
+The flow actions specified by :data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION` don't need a flow change
+confirmation event handler:
+
+* `'enter'` or `'leave'` extend/reduce the flow path.
+* `'focus'` pass/change the input focus.
+* `'suggest'` for autocompletion or other suggestions.
+
+All other flow actions need to confirmed for to be changed by :meth:`~MainAppBase.change_flow`, either by
+a custom flow change confirmation method/event-handler or by declaring a popup class that has the same name
+returned by :func:`flow_popup_class_name` if called with the new flow id.
+
+The name of the flow change confirmation method that gets fired when the app want to change the flow (via the method
 :meth:`~MainAppBase.change_flow`) depends on the new flow id and gets determined by the function
-:func:`flow_event_name`.
+:func:`flow_change_confirmation_event_name`.
 
-If the event handler exists and does return `True` for to confirm the flow change then flow id and path will be changed
-accordingly.
+If the flow-specific change confirmation event handler does not exist or returns in a boolean `False` or `None`
+then :meth:`~MainAppBase.on_flow_change` will be called. And if this call also returns `False` then the action
+of the new flow id will be searched within :data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION` and if not found
+then the flow change will be rejected and :meth:`~MainAppBase.change_flow` returns `False`
 
-If the flow-specific event call results in a boolean `False` or `None` then a more generic event
-:meth:`~MainAppBase.on_flow_change` will be called.
+If in contrary either the flow change confirmation event handler exists and does return `True` or
+:meth:`~MainAppBase.on_flow_change` returns True or the flow action of the new flow id is in
+:data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION` then the flow id and path will be changed accordingly.
 
-If also the generic event returns `False` then the action of the new flow id will be searched within
-:data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION` and if not found then the flow change will be rejected and
-:meth:`~MainAppBase.change_flow` returns `False`
+After the flow id/path change confirmation the method :meth:`~MainAppBase.change_flow` checks if the
+optional event_kwargs key `changed_event_name` got specified and if yes then it calls this method.
 
-If a confirmed flow change is resulting in a `'focus'` flow action then the event `on_flow_widget_focused` will fired
-and can be used by the GUI framework for to set the focus to the widget associated to the focus flow id.
+Finally if a confirmed flow change is resulting in a `'focus'` flow action then the event `on_flow_widget_focused`
+will be fired. This event can be used by the GUI framework for to set the focus to the widget associated with
+the new focus flow id.
 
 
 key press events
@@ -302,7 +316,7 @@ from abc import ABC, abstractmethod
 import re
 import traceback
 from copy import deepcopy
-from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type
 
 from ae.base import NAME_PARTS_SEP, snake_to_camel                  # type: ignore
 from ae.updater import check_all                                    # type: ignore
@@ -313,7 +327,7 @@ from ae.core import DEBUG_LEVELS                                    # type: igno
 from ae.console import ConsoleApp                                   # type: ignore
 
 
-__version__ = '0.1.42'
+__version__ = '0.1.45'
 
 
 AppStateType = Dict[str, Any]                   #: app state config variable type
@@ -409,8 +423,8 @@ def flow_key_split(flow_id: str) -> Tuple[str, str]:
     return flow_id, ""
 
 
-def flow_event_name(flow_id: str) -> str:
-    """ determine the name of the event method for the passed flow_id.
+def flow_change_confirmation_event_name(flow_id: str) -> str:
+    """ determine the name of the event method for the change confirmation of the passed flow_id.
 
     :param flow_id:         flow id.
     :return:                tuple with 2 items containing the flow action and the object name (and id).
@@ -535,38 +549,38 @@ class MainAppBase(ConsoleApp, ABC):
 
     def load_app_states(self):
         """ load application state from the config files for to prepare app.run_app """
-        app_state = dict()
+        app_states = dict()
         for key in self.app_state_keys():
-            app_state[key] = self.get_var(key, section=APP_STATE_SECTION_NAME)
+            app_states[key] = self.get_var(key, section=APP_STATE_SECTION_NAME)
 
-        self.setup_app_states(app_state)
+        self.setup_app_states(app_states)
 
     def on_app_init(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_init default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_init default/fallback flow change confirmation event handler called")
 
     def on_app_start(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_start default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_start default/fallback flow change confirmation event handler called")
 
     def on_app_run(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_run default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_run default/fallback flow change confirmation event handler called")
 
     def on_app_stop(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_stop default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_stop default/fallback flow change confirmation event handler called")
 
     def on_app_exit(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_exit default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_exit default/fallback flow change confirmation event handler called")
 
     def on_app_quit(self):
-        """ default/fallback event handler. """
-        self.vpo("MainAppBase.on_app_quit default/fallback event handler called")
+        """ default/fallback flow change confirmation event handler. """
+        self.vpo("MainAppBase.on_app_quit default/fallback flow change confirmation event handler called")
 
     def on_debug_level_change(self, debug_level: str, _event_kwargs: Dict[str, Any]) -> bool:
-        """ language app state change event handler.
+        """ debug level app state change flow change confirmation event handler.
 
         :param debug_level:     the new debug level name to be set (passed as flow key).
         :param _event_kwargs:   unused event kwargs.
@@ -578,7 +592,7 @@ class MainAppBase(ConsoleApp, ABC):
         return True
 
     def on_lang_code_change(self, lang_code: str, _event_kwargs: Dict[str, Any]) -> bool:
-        """ language app state change event handler.
+        """ language app state change flow change confirmation event handler.
 
         :param lang_code:       the new language code to be set (passed as flow key).
         :param _event_kwargs:   unused event kwargs.
@@ -592,7 +606,7 @@ class MainAppBase(ConsoleApp, ABC):
         return True
 
     def on_light_theme_change(self, _flow_key: str, event_kwargs: Dict[str, Any]) -> bool:
-        """ app theme app state change event handler.
+        """ app theme app state change flow change confirmation event handler.
 
         :param _flow_key:       flow key.
         :param event_kwargs:    event kwargs with key `'light_theme'` containing True|False for light|dark theme.
@@ -606,7 +620,7 @@ class MainAppBase(ConsoleApp, ABC):
     def retrieve_app_states(self) -> AppStateType:
         """ determine the state of a running app from the main app instance and return it as dict.
 
-        :return:    dict with all app states available in the config files.
+        :return:                dict with all app states available in the config files.
         """
         app_state = dict()
         for key in self.app_state_keys():
@@ -618,7 +632,7 @@ class MainAppBase(ConsoleApp, ABC):
     def save_app_states(self) -> str:
         """ save app state in config file.
 
-        :return:    empty string if app status could be saved into config files else error message.
+        :return:                empty string if app status could be saved into config files else error message.
         """
         err_msg = ""
 
@@ -647,13 +661,13 @@ class MainAppBase(ConsoleApp, ABC):
             self.play_sound('error' if err_msg else 'debug_save')
         return err_msg
 
-    def setup_app_states(self, app_state: AppStateType):
+    def setup_app_states(self, app_states: AppStateType):
         """ put app state variables into main app instance for to prepare framework app.run_app.
 
-        :param app_state:   dict of app states.
+        :param app_states:      dict of app states.
         """
-        self.vpo(f"MainAppBase.setup_app_states {app_state}")
-        for key, val in app_state.items():
+        self.vpo(f"MainAppBase.setup_app_states {app_states}")
+        for key, val in app_states.items():
             self.change_app_state(key, val, send_event=False)   # don't send on_-events as framework is not initialized
             if key == 'flow_id' and flow_action(val) == 'focus':
                 self._last_focus_flow_id = val
@@ -663,14 +677,14 @@ class MainAppBase(ConsoleApp, ABC):
             self.change_app_state(upd_key, upd_val, send_event=False)
             self.set_var(upd_key, upd_val, section=APP_STATE_SECTION_NAME, old_name=old_name)
 
-        config_file_version = app_state.get(APP_STATE_VERSION_VAR_NAME, 0)
+        config_file_version = app_states.get(APP_STATE_VERSION_VAR_NAME, 0)
         for version in range(config_file_version, APP_STATE_CURRENT_VERSION):
             if version == 0:
                 upd_app_state('light_theme', True)
             elif version == 1:
                 upd_app_state('sound_volume', 0.96)
             elif version == 2:
-                if 'context_id' not in app_state:
+                if 'context_id' not in app_states:
                     continue    # update - only needed for LiszApp apps
 
                 def mig_node(parent_node):
@@ -682,9 +696,9 @@ class MainAppBase(ConsoleApp, ABC):
                             item['node'] = mig_node(node)
                         new_parent_node.append(item)
                     return new_parent_node
-                upd_app_state('root_node', mig_node(app_state['data_tree']), 'data_tree')
-                upd_app_state('flow_path_ink', app_state['context_path_ink'], 'context_path_ink')
-                upd_app_state('flow_id_ink', app_state['context_id_ink'], 'context_id_ink')
+                upd_app_state('root_node', mig_node(app_states['data_tree']), 'data_tree')
+                upd_app_state('flow_path_ink', app_states['context_path_ink'], 'context_path_ink')
+                upd_app_state('flow_id_ink', app_states['context_id_ink'], 'context_id_ink')
                 upd_app_state('flow_path', [], 'context_path')
                 upd_app_state('flow_id', '', 'context_id')
 
@@ -704,45 +718,61 @@ class MainAppBase(ConsoleApp, ABC):
         """
         return stack_var(class_name)
 
-    def change_flow(self, new_flow_id: str, popups_to_close: Sequence = (), **event_kwargs) -> bool:
-        """ change/switch flow id.
+    def change_flow(self, new_flow_id: str, **event_kwargs) -> bool:
+        """ try to change/switch the current flow id to the value passed in :paramref:`~change_flow.new_flow_id`.
 
-        :param new_flow_id:     new flow id (maybe overwritten by event handlers in event_kwargs['flow_id']).
-        :param popups_to_close: optional sequence of widgets to be closed on confirmed flow change.
-        :param event_kwargs:    optional args to pass additional data or info onto and from the event handler.
+        :param new_flow_id:     new flow id (maybe overwritten by flow change confirmation event handlers by
+                                assigning a flow id to event_kwargs['flow_id']).
 
-                                info pass onto event handler (e.g. via :attr:`~ae.kivy_app.FlowButton.tap_kwargs`
-                                or :attr:`~ae.kivy_app.FlowToggler.tap_kwargs`):
+        :param event_kwargs:    optional args to pass additional data or info onto and from the flow change
+                                confirmation event handler.
 
-                                * `popup_kwargs`: optional dict passed to the Popup `__init__` method,
-                                  like e.g. dict(parent=parent_widget_of_popup, data=...).
-                                * `tap_widget`: optional tapped button widget instance (initiating this
-                                  flow change).
+                                The following keys are currently supported/implemented by this module/portion
+                                (additional keys can be added by the modules/apps using this method):
+
+                                * `changed_event_name`: optional main app event method name to be called
+                                  if the flow got confirmed and changed.
+                                * `count`: optional number used for to render a pluralized help text
+                                  for this flow change (for this gets also passed to the help text renderer by/in
+                                  :meth:`~ae.gui_help.HelpAppBase.change_flow`).
                                 * `edit_widget`: optional widget instance for edit/input.
-
-                                info passed from the event handler:
-
                                 * `flow_id`: process :attr:`~MainAppBase.flow_path` as specified by the
                                   :paramref:`~change_flow.new_flow_id` argument, but then overwrite this
                                   flow id with this event arg value for to set :attr:`~MainAppBase.flow_id`.
+                                * `popup_kwargs`: optional dict passed to the Popup `__init__` method,
+                                  like e.g. dict(parent=parent_widget_of_popup, data=...).
+                                * `popups_to_close`: optional sequence of widgets to be closed by this
+                                  method after flow change got confirmed. For to prevent weakref-errors
+                                  the `attach_to` attribute of each widget will additionally be reset to None.
+                                * 'reset_last_focus_flow_id': pass `True` for to reset the last focus flow id
+                                  to an empty flow id, pass `False` or `None` for to ignore the last focus id
+                                  (and not use for to set flow id) or pass a flow id string value
+                                  for to change the last focus flow id to the passed value.
+                                * `tap_widget`: optional tapped button widget instance (initiating this
+                                  flow change).
+
+                                Some of these keys get passed directly on the call of this method, e.g.
+                                via :attr:`~ae.kivy_app.FlowButton.tap_kwargs` or
+                                :attr:`~ae.kivy_app.FlowToggler.tap_kwargs`, where others get added by the
+                                flow change confirmation handlers/callbacks.
+
+        :return:                True if flow got confirmed by a declared custom flow change confirmation event handler
+                                (either event method or Popup class) of the app and changed accordingly, else False.
 
-        :return:                True if flow changed and got confirmed by a declared custom event handler
-                                (either event method or Popup class) of the app, else False.
                                 Some flow actions are handled internally independent from the
-                                return value of a found/declared
-                                custom event handler, like e.g. `'enter'` or `'leave'` will always
-                                extend/reduce the flow path and the action `'focus'` will give the
+                                return value of a found/declared custom event handler, like e.g. `'enter'` or `'leave'`
+                                will always extend/reduce the flow path and the action `'focus'` will give the
                                 indexed widget the input focus (these exceptions are configurable via
                                 the list :data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION`).
         """
-        self.vpo(f"MainAppBase.change_flow('{new_flow_id}', {popups_to_close}, {event_kwargs})"
-                 f" flow='{self.flow_id}' path={self.flow_path}")
+        self.vpo(f"MainAppBase.change_flow '{new_flow_id}'/'{self.flow_id}' path={self.flow_path} kwa={event_kwargs}")
 
         action = flow_action(new_flow_id)
-        if not self.call_method(flow_event_name(new_flow_id), flow_key(new_flow_id), event_kwargs) \
+        if not self.call_method(flow_change_confirmation_event_name(new_flow_id), flow_key(new_flow_id), event_kwargs) \
                 and not self.on_flow_change(new_flow_id, event_kwargs) \
                 and action not in ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION:
-            self.vpo(f"MainAppBase.change_flow REJECTED '{new_flow_id}'; flow='{self.flow_id}' path={self.flow_path}")
+            self.vpo(f"MainAppBase.change_flow REJECTED '{new_flow_id}'/'{self.flow_id}' path={self.flow_path}"
+                     f" kwa={event_kwargs}")
             return False
 
         has_flow_focus = flow_action(self.flow_id) == 'focus'
@@ -769,21 +799,33 @@ class MainAppBase(ConsoleApp, ABC):
         else:
             flow_id = new_flow_id if action == 'focus' else (self.flow_id if has_flow_focus else empty_flow)
 
-        popups_to_close = event_kwargs.get('popups_to_close', popups_to_close)
+        popups_to_close = event_kwargs.get('popups_to_close', ())
         for widget in reversed(popups_to_close):
             widget.close()
+            widget.attach_to = None     # mainly for DropDown widgets to prevent weakref-error/-exception
 
         if action not in ACTIONS_REDUCING_FLOW_PATH or not has_flow_focus:
-            flow_id = event_kwargs.get('flow_id', flow_id)  # update flow_id from event_kwargs set by event handler
-        if flow_id == empty_flow and self._last_focus_flow_id and action not in ACTIONS_EXTENDING_FLOW_PATH:
+            flow_id = event_kwargs.get('flow_id', flow_id)  # update flow_id from event_kwargs
+        if 'reset_last_focus_flow_id' in event_kwargs:
+            last_flow_id = event_kwargs['reset_last_focus_flow_id']
+            if last_flow_id is True:
+                self._last_focus_flow_id = empty_flow
+            elif isinstance(last_flow_id, str):
+                self._last_focus_flow_id = last_flow_id
+        elif flow_id == empty_flow and self._last_focus_flow_id and action not in ACTIONS_EXTENDING_FLOW_PATH:
             flow_id = self._last_focus_flow_id
         self.change_app_state('flow_id', flow_id)
+
+        changed_event_name = event_kwargs.get('changed_event_name', '')
+        if changed_event_name:
+            self.call_method(changed_event_name)
+
         if flow_action(flow_id) == 'focus':
             self.call_method('on_flow_widget_focused')
             self._last_focus_flow_id = flow_id
 
-        self.vpo(f"MainAppBase.change_flow CHANGED '{new_flow_id}';"
-                 f" flow='{self.flow_id}' last='{self._last_focus_flow_id}' path={self.flow_path}")
+        self.vpo(f"MainAppBase.change_flow CHANGED '{new_flow_id}'/'{self.flow_id}'/'{self._last_focus_flow_id}'"
+                 f" path={self.flow_path} kwa={event_kwargs}")
 
         return True
 
@@ -819,7 +861,7 @@ class MainAppBase(ConsoleApp, ABC):
                                 will be passed onto the `__init__` method of the found Popup class.
         :return:                True if Popup class was found and displayed.
 
-        This method is mainly used as the last fallback clicked event handler of a FlowButton.
+        This method is mainly used as the last fallback clicked flow change confirmation event handler of a FlowButton.
         """
         class_name = flow_popup_class_name(flow_id)
         self.vpo(f"MainAppBase.on_flow_change '{flow_id}' {event_kwargs} lookup={class_name}")
